@@ -1,6 +1,6 @@
 /* pkl.c - Poke compiler.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022, 2023 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -57,7 +57,11 @@
    LEXICAL_CUCKOLDING_P is 1 if alien tokens are to be recognized.
 
    ALIEN_TOKEN_FN is the user-provided handler for alien tokens.  This
-   field is NULL if the user didn't register a handler.  */
+   field is NULL if the user didn't register a handler.
+
+   ALIEN_DTOKEN_FN is the user-provided handler for delimited alien
+   tokens.  This field is NULL if the user didn't register a
+   handler.  */
 
 struct pkl_compiler
 {
@@ -67,11 +71,12 @@ struct pkl_compiler
   int compiling;
   int error_on_warning;
   int quiet_p;
+  int debug_p;
+  char *last_ast_str;
 #define PKL_MODULES_STEP 8
-  char **modules;
-  int num_modules;
   int lexical_cuckolding_p;
   pkl_alien_token_handler_fn alien_token_fn;
+  pkl_alien_dtoken_handler_fn alien_dtoken_fn;
 };
 
 
@@ -122,9 +127,9 @@ pkl_new (pvm vm, const char *rt_path,
   /* Be verbose by default :) */
   compiler->quiet_p = 0;
 
-  /* No modules loaded initially.  */
-  compiler->modules = NULL;
-  compiler->num_modules = 0;
+  /* No debug by default.  */
+  compiler->debug_p = 0;
+  compiler->last_ast_str = NULL;
 
   /* Bootstrap the compiler.  An error bootstraping is an internal
      error and should be reported as such.  */
@@ -185,12 +190,8 @@ out_of_memory:
 void
 pkl_free (pkl_compiler compiler)
 {
-  size_t i;
-
   pkl_env_free (compiler->env);
-  for (i = 0; i < compiler->num_modules; ++i)
-    free (compiler->modules[i]);
-  free (compiler->modules);
+  free (compiler->last_ast_str);
   free (compiler);
 }
 
@@ -304,6 +305,12 @@ rest_of_compilation (pkl_compiler compiler,
   if (analf_payload.errors > 0)
     goto error;
 
+  if (compiler->debug_p)
+    {
+      free (compiler->last_ast_str);
+      compiler->last_ast_str = pkl_ast_format (ast->ast);
+    }
+
   pkl_ast_free (ast);
   return gen_payload.program;
 
@@ -641,6 +648,24 @@ pkl_set_quiet_p (pkl_compiler compiler, int quiet_p)
   compiler->quiet_p = quiet_p;
 }
 
+int
+pkl_debug_p (pkl_compiler compiler)
+{
+  return compiler->debug_p;
+}
+
+void
+pkl_set_debug_p (pkl_compiler compiler, int debug_p)
+{
+  compiler->debug_p = debug_p;
+}
+
+char *
+pkl_get_last_ast_str (pkl_compiler compiler)
+{
+  return compiler->last_ast_str;
+}
+
 int
 pkl_lexical_cuckolding_p (pkl_compiler compiler)
 {
@@ -660,6 +685,12 @@ pkl_alien_token_fn (pkl_compiler compiler)
   return compiler->alien_token_fn;
 }
 
+pkl_alien_dtoken_handler_fn
+pkl_alien_dtoken_fn (pkl_compiler compiler)
+{
+  return compiler->alien_dtoken_fn;
+}
+
 void
 pkl_set_alien_token_fn (pkl_compiler compiler,
                         pkl_alien_token_handler_fn cb)
@@ -667,6 +698,13 @@ pkl_set_alien_token_fn (pkl_compiler compiler,
   compiler->alien_token_fn = cb;
 }
 
+void
+pkl_set_alien_dtoken_fn (pkl_compiler compiler,
+                         pkl_alien_dtoken_handler_fn cb)
+{
+  compiler->alien_dtoken_fn = cb;
+}
+
 pvm_program
 pkl_compile_call (pkl_compiler compiler, pvm_val cls, pvm_val *ret,
                   int narg, va_list ap)
@@ -699,41 +737,6 @@ pkl_get_vm (pkl_compiler compiler)
   return compiler->vm;
 }
 
-void
-pkl_add_module (pkl_compiler compiler, const char *path)
-{
-  const char *module = last_component (path);
-
-  if (pkl_module_loaded_p (compiler, path))
-    return;
-
-  if (compiler->num_modules % PKL_MODULES_STEP == 0)
-    {
-      size_t size = ((compiler->num_modules + PKL_MODULES_STEP)
-                     * sizeof (char*));
-      compiler->modules = realloc (compiler->modules, size);
-      memset (compiler->modules + compiler->num_modules, 0,
-              PKL_MODULES_STEP * sizeof (char*));
-    }
-
-  compiler->modules[compiler->num_modules++] = strdup (module);
-}
-
-int
-pkl_module_loaded_p (pkl_compiler compiler, const char *path)
-{
-  const char *basename = last_component (path);
-  int i;
-
-  for (i = 0; i < compiler->num_modules; ++i)
-    {
-      if (STREQ (compiler->modules[i], basename))
-        return 1;
-    }
-
-  return 0;
-}
-
 char *
 pkl_resolve_module (pkl_compiler compiler,
                     const char *module,
@@ -793,25 +796,15 @@ pkl_resolve_module (pkl_compiler compiler,
 }
 
 int
-pkl_load (pkl_compiler compiler, const char *module)
+pkl_load (pkl_compiler compiler, const char *module, pvm_val *exit_exception)
 {
-  pvm_val exit_exception;
   char *module_filename = pkl_resolve_module (compiler,
                                               module,
                                               0 /* filename_p */);
+
   if (!module_filename)
     return 0;
-
-  if (pkl_module_loaded_p (compiler, module_filename))
-    return 1;
-
-  if (!pkl_execute_file (compiler, module_filename,
-                         &exit_exception)
-      || exit_exception != PVM_NULL)
-    return 0;
-
-  pkl_add_module (compiler, module_filename);
-  return 1;
+  return pkl_execute_file (compiler, module_filename, exit_exception);
 }
 
 static pkl_ast_node
@@ -845,8 +838,10 @@ pvm_type_to_ast_type (pkl_ast ast, pvm_val type)
           = pvm_type_to_ast_type (ast, PVM_VAL_TYP_O_BASE_TYPE (type));
         pkl_ast_node unit
           = pkl_ast_make_integer (ast, PVM_VAL_ULONG (PVM_VAL_TYP_O_UNIT (type)));
+        pkl_ast_node ref_type
+          = pvm_type_to_ast_type (ast, PVM_VAL_TYP_O_REF_TYPE (type));
 
-        return pkl_ast_make_offset_type (ast, base_type, unit);
+        return pkl_ast_make_offset_type (ast, base_type, unit, ref_type);
         break;
       }
     case PVM_TYPE_VOID:
