@@ -1,6 +1,6 @@
 /* poke.c - Interactive editor for binary files.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022, 2023 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -107,6 +107,11 @@ int poke_default_auto_map_p = 1;
 
 int poke_no_hserver_arg = 0;
 
+/* The following global determines whether the user specified to not
+   load "standard" types in the Poke incremental compiler.  */
+
+int poke_no_stdtypes_arg = 0;
+
 /* This is used by commands to indicate to the REPL that it must
    exit.  */
 
@@ -138,7 +143,8 @@ enum
   STYLE_DARK_ARG,
   STYLE_BRIGHT_ARG,
   NO_AUTO_MAP_ARG,
-  NO_HSERVER_ARG
+  NO_HSERVER_ARG,
+  NO_STDTYPES_ARG
 };
 
 static const struct option long_options[] =
@@ -156,6 +162,7 @@ static const struct option long_options[] =
   {"style-bright", no_argument, NULL, STYLE_BRIGHT_ARG},
   {"no-auto-map", no_argument, NULL, NO_AUTO_MAP_ARG},
   {"no-hserver", no_argument, NULL, NO_HSERVER_ARG},
+  {"no-stdtypes", no_argument, NULL, NO_STDTYPES_ARG},
   {NULL, 0, NULL, 0},
 };
 
@@ -202,11 +209,18 @@ print_help (void)
 #if HAVE_HSERVER
   puts (_("      --no-hserver                    do not run the hyperlinks server"));
 #endif
+  puts (_("      --no-stdtypes                   do not define standard types"));
   puts (_("      --quiet                         be as terse as possible"));
   puts (_("      --help                          print a help message and exit"));
   puts (_("      --version                       show version and exit"));
 
   puts ("");
+  puts (_("The following environment variables, if set, are used by poke:"));
+  puts ("");
+  puts (_("      POKE_LOAD_PATH                  List of file paths separated"));
+  puts (_("                                      by colon characters (:) which"));
+  puts (_("                                      is prepended to the load_path"));
+  puts (_("                                      when poke starts"));
   /* TRANSLATORS: --help output 5+ (reports)
      TRANSLATORS: the placeholder indicates the bug-reporting address
      for this application.  Please add _another line_ with the
@@ -272,6 +286,7 @@ finalize (void)
     pk_hserver_shutdown ();
 #endif
   pk_cmd_shutdown ();
+  pk_ios_shutdown ();
   pk_map_shutdown ();
   pk_compiler_free (poke_compiler);
   pk_term_shutdown ();
@@ -281,18 +296,18 @@ finalize (void)
 
 static struct pk_term_if poke_term_if =
   {
-    .flush_fn = pk_term_flush,
-    .puts_fn = pk_puts,
-    .printf_fn = pk_printf,
-    .indent_fn = pk_term_indent,
-    .class_fn = pk_term_class,
-    .end_class_fn = pk_term_end_class,
-    .hyperlink_fn = pk_term_hyperlink,
-    .end_hyperlink_fn = pk_term_end_hyperlink,
-    .get_color_fn = pk_term_get_color,
-    .set_color_fn = pk_term_set_color,
-    .get_bgcolor_fn = pk_term_get_bgcolor,
-    .set_bgcolor_fn = pk_term_set_bgcolor,
+    .flush_fn = pk_term_flush_1,
+    .puts_fn = pk_puts_1,
+    .printf_fn = pk_printf_1,
+    .indent_fn = pk_term_indent_1,
+    .class_fn = pk_term_class_1,
+    .end_class_fn = pk_term_end_class_1,
+    .hyperlink_fn = pk_term_hyperlink_1,
+    .end_hyperlink_fn = pk_term_end_hyperlink_1,
+    .get_color_fn = pk_term_get_color_1,
+    .set_color_fn = pk_term_set_color_1,
+    .get_bgcolor_fn = pk_term_get_bgcolor_1,
+    .set_bgcolor_fn = pk_term_set_bgcolor_1,
   };
 
 const char *
@@ -317,7 +332,7 @@ void
 pk_set_var_int (const char *name, int value)
 {
   pk_decl_set_val (poke_compiler, name,
-                   pk_make_int (value, 32));
+                   pk_make_int (poke_compiler, value, 32));
 }
 
 static void
@@ -342,13 +357,14 @@ set_script_args (int argc, char *argv[])
   /* Any argument after SCRIPT is an argument for the script.  */
   i = i + 2;
   nargs = argc - i;
-  argv_array = pk_make_array (pk_make_uint (nargs, 64),
-                              pk_make_array_type (pk_make_string_type (),
-                                                  array_type_bounder));
+  argv_array = pk_make_array (
+      poke_compiler, pk_make_uint (poke_compiler, nargs, 64),
+      pk_make_array_type (poke_compiler, pk_make_string_type (poke_compiler),
+                          array_type_bounder));
 
   for (index = 0; i < argc; ++i, ++index)
     pk_array_insert_elem (argv_array, index,
-                          pk_make_string (argv[i]));
+                          pk_make_string (poke_compiler, argv[i]));
 
   pk_defvar (poke_compiler, "argv", argv_array);
 }
@@ -385,6 +401,9 @@ parse_args_1 (int argc, char *argv[])
         case NO_HSERVER_ARG:
           poke_no_hserver_arg = 1;
           break;
+        case NO_STDTYPES_ARG:
+          poke_no_stdtypes_arg = 1;
+          break;
         case 'q':
         case NO_INIT_FILE_ARG:
           poke_load_init_file = 0;
@@ -488,6 +507,7 @@ parse_args_2 (int argc, char *argv[])
           }
         case NO_AUTO_MAP_ARG:
         case NO_HSERVER_ARG:
+        case NO_STDTYPES_ARG:
         case 'q':
         case NO_INIT_FILE_ARG:
           /* These are handled in parse_args_1.  */
@@ -593,13 +613,15 @@ initialize (int argc, char *argv[])
   pk_term_init (argc, argv);
 
   /* Initialize the poke incremental compiler.  */
-  poke_compiler = pk_compiler_new (&poke_term_if);
+  poke_compiler = pk_compiler_new_with_flags (&poke_term_if,
+                                              poke_no_stdtypes_arg ? PK_F_NOSTDTYPES : 0);
   if (poke_compiler == NULL)
     pk_fatal ("creating the incremental compiler");
 
   /* Make poke_interactive_p available to poke programs.  */
   if (pk_defvar (poke_compiler, "poke_interactive_p",
-                 pk_make_int (poke_interactive_p, 32)) == PK_ERROR)
+                 pk_make_int (poke_compiler, poke_interactive_p, 32))
+      == PK_ERROR)
     pk_fatal ("defining poke_interactive_p");
 
   /* Add the directories where poke installs .pk files to the
@@ -607,9 +629,18 @@ initialize (int argc, char *argv[])
   {
     pk_val load_path = pk_decl_val (poke_compiler, "load_path");
     char *newpaths;
+    const char *env_load_path = getenv ("POKE_LOAD_PATH");
+    char *user_load_path = NULL;
+
+    if (env_load_path)
+      {
+        user_load_path = pk_str_concat (env_load_path, ":", NULL);
+        pk_assert_alloc (user_load_path);
+      }
 
     assert (load_path != PK_NULL);
-    newpaths = pk_str_concat (pk_string_str (load_path),
+    newpaths = pk_str_concat (user_load_path ? user_load_path : "",
+                              pk_string_str (load_path),
                               ":",
                               poke_appdir,
                               ":",
@@ -618,19 +649,27 @@ initialize (int argc, char *argv[])
     pk_assert_alloc (newpaths);
 
     pk_decl_set_val (poke_compiler,
-                     "load_path", pk_make_string (newpaths));
+                     "load_path", pk_make_string (poke_compiler, newpaths));
     free (newpaths);
   }
 
-  /* The Poke hyperlinks facilities must be loaded before poke.pk and
-     the cmd subsystem.  This is done even if the hserver is
-     disabled.  */
-  if (pk_load (poke_compiler, "pk-hserver") != PK_OK)
-    pk_fatal ("unable to load the pk-hserver module");
+  {
+    pk_val exception;
 
-  /* Load poke.pk  */
-  if (pk_load (poke_compiler, "poke") != PK_OK)
-    pk_fatal ("unable to load the poke module");
+    /* The Poke hyperlinks facilities must be loaded before poke.pk and
+       the cmd subsystem.  This is done even if the hserver is
+       disabled.  */
+    if (pk_load (poke_compiler, "pk-hserver", &exception) != PK_OK
+        || exception != PK_NULL)
+      /* There's no default exception handler yet to handle `exception'.  */
+      pk_fatal ("unable to load the pk-hserver module");
+
+    /* Load poke.pk  */
+    if (pk_load (poke_compiler, "poke", &exception) != PK_OK
+        || exception != PK_NULL)
+      /* There's no default exception handler yet to handle `exception'.  */
+      pk_fatal ("unable to load the poke module");
+  }
 
   /* Set the values of a few global variables defined in poke.pk.  */
   {
@@ -667,13 +706,13 @@ initialize (int argc, char *argv[])
     pk_decl_set_val (poke_compiler, "pk_network_endian", network_endian);
 
     pk_decl_set_val (poke_compiler, "pk_auto_map_p",
-                     pk_make_int (poke_default_auto_map_p, 32));
+                     pk_make_int (poke_compiler, poke_default_auto_map_p, 32));
 
     pk_decl_set_val (poke_compiler, "pk_have_libnbd_p",
 #if defined HAVE_LIBNBD
-                     pk_make_int (1, 32)
+                     pk_make_int (poke_compiler, 1, 32)
 #else
-                     pk_make_int (0, 32)
+                     pk_make_int (poke_compiler, 0, 32)
 #endif
                      );
   }
@@ -681,6 +720,9 @@ initialize (int argc, char *argv[])
   /* Initialize the global map.  */
   pk_map_init ();
 
+  /* Initialize ios stuff.  */
+  pk_ios_init ();
+
   /* Initialize the command subsystem.  This should be done even if
      called non-interactively.  */
   pk_cmd_init ();
@@ -699,7 +741,7 @@ initialize (int argc, char *argv[])
 
   /* Set the value of the Poke variable pk_hserver_p.  */
   pk_decl_set_val (poke_compiler, "pk_hserver_p",
-                   pk_make_int (poke_hserver_p, 32));
+                   pk_make_int (poke_compiler, poke_hserver_p, 32));
 #endif
 }
 
